__all__ = [
    "__title__",
    "__summary__",
    "__uri__",
    "__author__",
    "__email__",
    "__license__",
    "__copyright__",
]

__title__ = "layered_config_tree"
__summary__ = (
    "Layered Config Tree is a configuration structure which supports cascading layers."
)
__uri__ = "https://github.com/ihmeuw/layered_config_tree"

__author__ = "The vivarium developers"
__email__ = "vivarium.dev@gmail.com"

__license__ = "BSD-3-Clause"
__copyright__ = f"Copyright 2016-2024 {__author__}"
